
#include <stdio.h>
#include <asoundlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <libgen.h>

#define MAXARGS 255
#define ARGLEN 255
char *args[MAXARGS];

char *clientname="midiscript";
char *portname=NULL;

int open_seq(snd_seq_t **seq_handle) {
	int in_port;

	if(snd_seq_open(seq_handle, "default", SND_SEQ_OPEN_INPUT, 0)<0) {
		fprintf(stderr, "Error opening ALSA sequencer.\n");
		return(-1);
	}

	snd_seq_set_client_name(*seq_handle, clientname);

	in_port=snd_seq_create_simple_port(*seq_handle, portname,
		SND_SEQ_PORT_CAP_WRITE|SND_SEQ_PORT_CAP_SUBS_WRITE,
		SND_SEQ_PORT_TYPE_APPLICATION);

	if(in_port<0) {
		fprintf(stderr, "Error creating sequencer port.\n");
		return -1;
	}
	return in_port;
}

void do_fork(int num_args) {
	pid_t pid;
	pid=fork();
	int status;
	if(pid==0) {
		pid=fork();
		if(pid==0) {
			args[num_args]=NULL;
			execvp(args[0], args);
			fprintf(stderr, "couldn't exec %s\n", args[0]);
			exit(0);
		}
		if(pid>0) exit(0);
	}
	if(pid>0) wait(&status);
}

void read_midi(snd_seq_t *seq_handle) {
	snd_seq_event_t *ev;

	do {
		snd_seq_event_input(seq_handle, &ev);
		int i=1;
		snprintf(args[i++], ARGLEN, "%d:%d",
			ev->source.client, ev->source.port);
		snprintf(args[i++], ARGLEN, "%d:%d",
			ev->dest.client, ev->source.port);
		switch(ev->type) {
			case SND_SEQ_EVENT_SYSTEM:
				snprintf(args[i++], ARGLEN, "SYSTEM");
				snprintf(args[i++], ARGLEN, "%d", ev->data.result.event);
				snprintf(args[i++], ARGLEN, "%d", ev->data.result.result);
				break;
			case SND_SEQ_EVENT_RESULT:
				snprintf(args[i++], ARGLEN, "RESULT");
				snprintf(args[i++], ARGLEN, "%d", ev->data.result.event);
				snprintf(args[i++], ARGLEN, "%d", ev->data.result.result);
				break;
			case SND_SEQ_EVENT_NOTE:
				snprintf(args[i++], ARGLEN, "NOTE");
				snprintf(args[i++], ARGLEN, "%d", ev->data.note.channel);
				snprintf(args[i++], ARGLEN, "%d", ev->data.note.note);
				snprintf(args[i++], ARGLEN, "%d", ev->data.note.velocity);
				snprintf(args[i++], ARGLEN, "%d", ev->data.note.off_velocity);
				snprintf(args[i++], ARGLEN, "%d", ev->data.note.duration);
				break;
			case SND_SEQ_EVENT_NOTEON:
				snprintf(args[i++], ARGLEN, "NOTEON");
				snprintf(args[i++], ARGLEN, "%d", ev->data.note.channel);
				snprintf(args[i++], ARGLEN, "%d", ev->data.note.note);
				snprintf(args[i++], ARGLEN, "%d", ev->data.note.velocity);
				snprintf(args[i++], ARGLEN, "%d", ev->data.note.off_velocity);
				snprintf(args[i++], ARGLEN, "%d", ev->data.note.duration);
				break;
			case SND_SEQ_EVENT_NOTEOFF:
				snprintf(args[i++], ARGLEN, "NOTEOFF");
				snprintf(args[i++], ARGLEN, "%d", ev->data.note.channel);
				snprintf(args[i++], ARGLEN, "%d", ev->data.note.note);
				snprintf(args[i++], ARGLEN, "%d", ev->data.note.velocity);
				snprintf(args[i++], ARGLEN, "%d", ev->data.note.off_velocity);
				snprintf(args[i++], ARGLEN, "%d", ev->data.note.duration);
				break;
			case SND_SEQ_EVENT_KEYPRESS:
				snprintf(args[i++], ARGLEN, "KEYPRESS");
				snprintf(args[i++], ARGLEN, "%d", ev->data.note.channel);
				snprintf(args[i++], ARGLEN, "%d", ev->data.note.note);
				snprintf(args[i++], ARGLEN, "%d", ev->data.note.velocity);
				snprintf(args[i++], ARGLEN, "%d", ev->data.note.off_velocity);
				snprintf(args[i++], ARGLEN, "%d", ev->data.note.duration);
				break;
			case SND_SEQ_EVENT_CONTROLLER:
				snprintf(args[i++], ARGLEN, "CONTROLLER");
				snprintf(args[i++], ARGLEN, "%d", ev->data.control.channel);
				snprintf(args[i++], ARGLEN, "%d", ev->data.control.param);
				snprintf(args[i++], ARGLEN, "%d", ev->data.control.value);
				break;
			case SND_SEQ_EVENT_PGMCHANGE:
				snprintf(args[i++], ARGLEN, "PGMCHANGE");
				snprintf(args[i++], ARGLEN, "%d", ev->data.control.channel);
				snprintf(args[i++], ARGLEN, "%d", ev->data.control.param);
				snprintf(args[i++], ARGLEN, "%d", ev->data.control.value);
				break;
			case SND_SEQ_EVENT_CHANPRESS:
				snprintf(args[i++], ARGLEN, "CHANPRESS");
				snprintf(args[i++], ARGLEN, "%d", ev->data.control.channel);
				snprintf(args[i++], ARGLEN, "%d", ev->data.control.param);
				snprintf(args[i++], ARGLEN, "%d", ev->data.control.value);
				break;
			case SND_SEQ_EVENT_PITCHBEND:
				snprintf(args[i++], ARGLEN, "PITCHBEND");
				snprintf(args[i++], ARGLEN, "%d", ev->data.control.channel);
				snprintf(args[i++], ARGLEN, "%d", ev->data.control.param);
				snprintf(args[i++], ARGLEN, "%d", ev->data.control.value);
				break;
			case SND_SEQ_EVENT_CONTROL14:
				break;
			case SND_SEQ_EVENT_NONREGPARAM:
				break;
			case SND_SEQ_EVENT_REGPARAM:
				break;
			case SND_SEQ_EVENT_SONGPOS:
				break;
			case SND_SEQ_EVENT_SONGSEL:
				break;
			case SND_SEQ_EVENT_QFRAME:
				break;
			case SND_SEQ_EVENT_TIMESIGN:
				break;
			case SND_SEQ_EVENT_KEYSIGN:
				break;
			case SND_SEQ_EVENT_START:
				snprintf(args[i++], ARGLEN, "START");
				snprintf(args[i++], ARGLEN, "%d", ev->data.queue.queue);
				snprintf(args[i++], ARGLEN, "%d", ev->data.queue.param.value);
				break;
			case SND_SEQ_EVENT_CONTINUE:
				snprintf(args[i++], ARGLEN, "CONTINUE");
				snprintf(args[i++], ARGLEN, "%d", ev->data.queue.queue);
				snprintf(args[i++], ARGLEN, "%d", ev->data.queue.param.value);
				break;
			case SND_SEQ_EVENT_STOP:
				snprintf(args[i++], ARGLEN, "STOP");
				snprintf(args[i++], ARGLEN, "%d", ev->data.queue.queue);
				snprintf(args[i++], ARGLEN, "%d", ev->data.queue.param.value);
				break;
			case SND_SEQ_EVENT_SETPOS_TICK:
				snprintf(args[i++], ARGLEN, "SETPOS_TICK");
				snprintf(args[i++], ARGLEN, "%d", ev->data.queue.queue);
				snprintf(args[i++], ARGLEN, "%d", ev->data.queue.param.value);
				break;
			case SND_SEQ_EVENT_SETPOS_TIME:
				snprintf(args[i++], ARGLEN, "SETPOS_TIME");
				snprintf(args[i++], ARGLEN, "%d", ev->data.queue.queue);
				snprintf(args[i++], ARGLEN, "%d", ev->data.queue.param.value);
				break;
			case SND_SEQ_EVENT_TEMPO:
				snprintf(args[i++], ARGLEN, "TEMPO");
				snprintf(args[i++], ARGLEN, "%d", ev->data.queue.queue);
				snprintf(args[i++], ARGLEN, "%d", ev->data.queue.param.value);
				break;
			case SND_SEQ_EVENT_CLOCK:
				snprintf(args[i++], ARGLEN, "CLOCK");
				snprintf(args[i++], ARGLEN, "%d", ev->data.queue.queue);
				snprintf(args[i++], ARGLEN, "%d", ev->data.queue.param.value);
				break;
			case SND_SEQ_EVENT_TICK:
				snprintf(args[i++], ARGLEN, "TICK");
				snprintf(args[i++], ARGLEN, "%d", ev->data.queue.queue);
				snprintf(args[i++], ARGLEN, "%d", ev->data.queue.param.value);
				break;
			case SND_SEQ_EVENT_QUEUE_SKEW:
				snprintf(args[i++], ARGLEN, "QUEUE_SKEW");
				snprintf(args[i++], ARGLEN, "%d", ev->data.queue.queue);
				snprintf(args[i++], ARGLEN, "%d", ev->data.queue.param.value);
				break;
			case SND_SEQ_EVENT_SYNC_POS:
				snprintf(args[i++], ARGLEN, "");
				snprintf(args[i++], ARGLEN, "%d", ev->data.queue.queue);
				snprintf(args[i++], ARGLEN, "%d", ev->data.queue.param.value);
				break;
			case SND_SEQ_EVENT_PORT_SUBSCRIBED:
				snprintf(args[i++], ARGLEN, "PORT_SUBSCRIBED");
				snprintf(args[i++], ARGLEN, "%d", ev->data.connect.sender);
				snprintf(args[i++], ARGLEN, "%d", ev->data.connect.dest);
				break;
			case SND_SEQ_EVENT_PORT_UNSUBSCRIBED:
				snprintf(args[i++], ARGLEN, "PORT_UNSUBSCRIBED");
				snprintf(args[i++], ARGLEN, "%d", ev->data.connect.sender);
				snprintf(args[i++], ARGLEN, "%d", ev->data.connect.dest);
				break;
			default:
				snprintf(args[i++], ARGLEN, "EVENT_UNKNOWN");
				break;
		}
		do_fork(i);
	} while (snd_seq_event_input_pending(seq_handle, 0)>0);
}

void usage(char *path) {
	printf("Usage: %s [-c client:port] [-n clientname] [-p portname] filename\n", path);
	exit(0);
}

int main(int argc, char **argv) {
	snd_seq_t *seq_handle;
	int in_port;
	int npfd;
	struct pollfd *pfd;

	char *connect_arg=NULL;

	if(argc<2) usage(argv[0]);

	int c;
	while((c=getopt(argc, argv, "c:hn:p:"))!=-1) {
		switch (c) {
			case 'c':
				connect_arg=optarg;
				break;
			case 'h':
				usage(argv[0]);
				break;
			case 'n':
				clientname=optarg;
				break;
			case 'p':
				portname=optarg;
				break;
		}
	}

	args[0]=argv[optind];
	int i;
	for(i=1; i<MAXARGS; i++) args[i]=malloc(ARGLEN);

	if(portname==NULL) portname=basename(argv[optind]);

	in_port=open_seq(&seq_handle);
	if (in_port<0) {
		fprintf(stderr, "ALSA Error.\n");
		exit(1);
	}

	if(connect_arg!=NULL) {
		int src_client;
		int src_port;
		if(sscanf(connect_arg, "%d:%d", &src_client, &src_port)!=2) {
			fprintf(stderr, "can't parse connection argument %s\n", connect_arg);
			exit(1);
		}
		printf("connecting to %d:%d...", src_client, src_port);
		if(snd_seq_connect_from(seq_handle, 0, src_client, src_port)) {
			printf("fail\n");
		} else {
			printf("success\n");
		}
	}

	npfd=snd_seq_poll_descriptors_count(seq_handle, POLLIN);
	pfd=(struct pollfd *)alloca(npfd * sizeof(struct pollfd));
	snd_seq_poll_descriptors(seq_handle, pfd, npfd, POLLIN);
	while (1) {
		if (poll(pfd, npfd, 100000) > 0) {
			read_midi(seq_handle);
		}  
	}
}

